from contextlib import closing
from os import path, makedirs
import fileinput
import json
import six

from pyramid.path import AssetResolver
from pyramid.settings import asbool, truthy
from pyramid.threadlocal import get_current_request
from webassets import Bundle
from webassets import __version__ as webassets_version
from webassets.env import Environment, Resolver
from webassets.exceptions import BundleError
from webassets.loaders import YAMLLoader
from zope.interface import Interface

USING_WEBASSETS_CONTEXT = webassets_version > (0, 9)

falsy = frozenset(('f', 'false', 'n', 'no', 'off', '0'))
booly = frozenset(list(truthy) + list(falsy))
auto_booly = frozenset(('true', 'false'))


def maybebool(value):
    '''
    If `value` is a string type, attempts to convert it to a boolean
    if it looks like it might be one, otherwise returns the value
    unchanged. The difference between this and
    :func:`pyramid.settings.asbool` is how non-bools are handled: this
    returns the original value, whereas `asbool` returns False.
    '''
    if isinstance(value, six.string_types) and value.lower() in booly:
        return asbool(value)  # pragma: no cover
    return value


def text(value):
    if type(value) is six.binary_type:
        return value.decode('utf-8')
    else:
        return value


class PyramidResolver(Resolver):
    def __init__(self):
        super(PyramidResolver, self).__init__()
        self.resolver = AssetResolver(None)

    def _split_spec(self, item):
        if ':' in item:
            package, subpath = item.split(':', 1)
            return (package, subpath)
        else:
            return (None, item)

    def _resolve_spec(self, spec):
        package, subpath = self._split_spec(spec)

        try:
            pkgpath = self.resolver.resolve(package + ':').abspath()
        except ImportError as e:
            raise BundleError(e)
        else:
            return path.join(pkgpath, subpath)

    def search_for_source(self, ctx, item):
        package, subpath = self._split_spec(item)
        if package is None:
            if USING_WEBASSETS_CONTEXT:
                return super(PyramidResolver, self).search_for_source(
                    ctx,
                    item
                )
            else:  # pragma: no cover
                return super(PyramidResolver, self).search_for_source(
                    item
                )
        else:
            pkgpath = self._resolve_spec(package + ':')
            return self.consider_single_directory(pkgpath, subpath)

    def resolve_source_to_url(self, ctx, filepath, item):
        request = get_current_request()

        # Use the filepath to reconstruct the item without globs
        package, _ = self._split_spec(item)
        if package is not None:
            pkgdir = self._resolve_spec(package + ':')
            if filepath.startswith(pkgdir):
                item = '{}:{}'.format(package, filepath[len(pkgdir):])

        # Attempt to resolve the filepath as passed (but after versioning).
        # If this fails, it may be because the static route was registered
        # with an asset spec. In this case, the original item may also be
        # an asset spec contained therein, so try to resolve that.
        if request is not None:
            for attempt in (filepath, item):
                try:
                    return request.static_url(attempt)
                except ValueError:
                    pass

        if USING_WEBASSETS_CONTEXT:
            return super(PyramidResolver, self).resolve_source_to_url(
                ctx,
                filepath,
                item
            )
        else:  # pragma: no cover
            return super(PyramidResolver, self).resolve_source_to_url(
                filepath,
                item
            )

    def resolve_output_to_path(self, ctx, target, bundle):
        package, filepath = self._split_spec(target)
        if package is not None:
            pkgpath = self._resolve_spec(package + ':')
            target = path.join(pkgpath, filepath)

        if USING_WEBASSETS_CONTEXT:
            return super(PyramidResolver, self).resolve_output_to_path(
                ctx,
                target,
                bundle
            )
        else:  # pragma: no cover
            return super(PyramidResolver, self).resolve_output_to_path(
                target,
                bundle
            )

    def resolve_output_to_url(self, ctx, item):
        request = get_current_request()

        if not path.isabs(item):
            if ':' not in item:
                if 'asset_base' in ctx.config:
                    if ctx.config['asset_base'].endswith(':'):
                        item = ctx.config['asset_base'] + item
                    else:
                        item = path.join(ctx.config['asset_base'], item)
                else:
                    item = path.join(ctx.directory, item)

        if ':' in item:
            filepath = self._resolve_spec(item)
        else:
            filepath = item

        if request is not None:
            for attempt in (filepath, item):
                try:
                    return request.static_url(item)
                except ValueError:
                    pass

        if USING_WEBASSETS_CONTEXT:
            return super(PyramidResolver, self).resolve_output_to_url(
                ctx,
                filepath
            )
        else:  # pragma: no cover
            return super(PyramidResolver, self).resolve_output_to_url(filepath)


class LegacyPyramidResolver(PyramidResolver):  # pragma: no cover
    def __init__(self, env):
        Resolver.__init__(self, env)
        self.resolver = AssetResolver(None)

    def search_for_source(self, *args):
        return PyramidResolver.search_for_source(self, self.env, *args)

    def resolve_source_to_url(self, *args):
        return PyramidResolver.resolve_source_to_url(self, self.env, *args)

    def resolve_output_to_path(self, *args):
        return PyramidResolver.resolve_output_to_path(self, self.env, *args)

    def resolve_output_to_url(self, *args):
        return PyramidResolver.resolve_output_to_url(self, self.env, *args)


class Environment(Environment):
    @property
    def resolver_class(self):
        if USING_WEBASSETS_CONTEXT:
            return PyramidResolver
        else:  # pragma: no cover
            return LegacyPyramidResolver


class IWebAssetsEnvironment(Interface):
    pass


def add_webasset(config, name, bundle):
    asset_env = get_webassets_env(config)
    asset_env.register(name, bundle)


def get_webassets_env(config):
    return config.registry.queryUtility(IWebAssetsEnvironment)


def get_webassets_env_from_settings(settings, prefix='webassets'):
    """This function will take all webassets.* parameters, and
    call the ``Environment()`` constructor with kwargs passed in.

    The only two parameters that are not passed as keywords are:

    * base_dir
    * base_url

    which are passed in positionally.

    Read the ``WebAssets`` docs for ``Environment`` for more details.
    """
    # Make a dictionary of the webassets.* elements...
    kwargs = {}   # assets settings
    cut_prefix = len(prefix) + 1
    for k in settings:
        if k.startswith(prefix):
            val = settings[k]
            if isinstance(val, six.string_types):
                if val.lower() in auto_booly:
                    val = asbool(val)
                elif val.lower().startswith('json:') and k[cut_prefix:] != 'manifest':
                    val = json.loads(val[5:])
            kwargs[k[cut_prefix:]] = val

    if 'base_dir' not in kwargs:
        raise Exception("You need to provide webassets.base_dir in your configuration")
    if 'base_url' not in kwargs:
        raise Exception("You need to provide webassets.base_url in your configuration")

    asset_dir = kwargs.pop('base_dir')
    asset_url = kwargs.pop('base_url')

    if ':' in asset_dir:
        try:
            resolved_dir = AssetResolver(None).resolve(asset_dir).abspath()
        except ImportError:
            pass
        else:
            # Store the original asset spec to use later
            kwargs['asset_base'] = asset_dir
            asset_dir = resolved_dir

    if not asset_url.startswith('/'):
        if six.moves.urllib.parse.urlparse(asset_url).scheme == '':
            asset_url = '/' + asset_url

    if 'debug' in kwargs:
        kwargs['debug'] = maybebool(kwargs['debug'])

    if 'cache' in kwargs:
        cache = kwargs['cache'] = maybebool(kwargs['cache'])

        if cache and isinstance(cache, six.string_types) and not path.isdir(cache):
            makedirs(cache)

    # 'updater' is just passed in...

    if 'auto_build' in kwargs:
        kwargs['auto_build'] = maybebool(kwargs['auto_build'])

    if 'jst_compiler' in kwargs:
        kwargs['JST_COMPILER'] = kwargs.pop('jst_compiler')

    if 'jst_namespace' in kwargs:
        kwargs['JST_NAMESPACE'] = kwargs.pop('jst_namespace')

    if 'manifest' in kwargs:
        kwargs['manifest'] = maybebool(kwargs['manifest'])

    if 'url_expire' in kwargs:
        kwargs['url_expire'] = maybebool(kwargs['url_expire'])

    if 'static_view' in kwargs:
        kwargs['static_view'] = asbool(kwargs['static_view'])
    else:
        kwargs['static_view'] = False

    if 'cache_max_age' in kwargs:
        kwargs['cache_max_age'] = int(kwargs.pop('cache_max_age'))
    else:
        kwargs['cache_max_age'] = None

    if 'load_path' in kwargs:
        # force load_path to be an array and split on whitespace
        if not isinstance(kwargs['load_path'], list):
            kwargs['load_path'] = kwargs['load_path'].split()

    paths = kwargs.pop('paths', None)

    if 'bundles' in kwargs:
        if isinstance(kwargs['bundles'], six.string_types):
            kwargs['bundles'] = kwargs['bundles'].split()

    bundles = kwargs.pop('bundles', None)

    assets_env = Environment(asset_dir, asset_url, **kwargs)

    if paths is not None:
        for map_path, map_url in json.loads(paths).items():
            assets_env.append_path(map_path, map_url)

    def yaml_stream(fname, mode):
        if path.exists(fname):
            return open(fname, mode)
        else:
            return assets_env.resolver.resolver.resolve(fname).stream()

    if isinstance(bundles, list):
        fnames = reversed(bundles)
        fin = fileinput.input(fnames, openhook=yaml_stream)
        with closing(fin):
            lines = [text(line).rstrip() for line in fin]
        yamlin = six.StringIO('\n'.join(lines))
        loader = YAMLLoader(yamlin)
        result = loader.load_bundles()
        assets_env.register(result)
    elif isinstance(bundles, dict):
        assets_env.register(bundles)

    return assets_env


def get_webassets_env_from_request(request):
    """ Get the webassets environment in the registry from the request. """
    return request.registry.queryUtility(IWebAssetsEnvironment)


def add_setting(config, key, value):
    env = config.registry.queryUtility(IWebAssetsEnvironment)
    env.config[key] = value


def add_path(config, path, url):
    env = config.registry.queryUtility(IWebAssetsEnvironment)
    env.append_path(path, url)


def assets(request, *args, **kwargs):
    env = get_webassets_env_from_request(request)

    result = []

    for f in args:
        try:
            result.append(env[f])
        except KeyError:
            result.append(f)

    bundle = Bundle(*result, **kwargs)
    if USING_WEBASSETS_CONTEXT:
        with bundle.bind(env):
            urls = bundle.urls()
    else:  # pragma: no cover
        urls = bundle.urls(env=env)

    return urls


def add_assets_global(event):
    event['webassets'] = assets


def includeme(config):
    config.add_subscriber(add_assets_global, 'pyramid.events.BeforeRender')

    settings = config.registry.settings
    assets_env = get_webassets_env_from_settings(settings)

    config.registry.registerUtility(assets_env, IWebAssetsEnvironment)

    config.add_directive('add_webasset', add_webasset)
    config.add_directive('get_webassets_env', get_webassets_env)
    config.add_directive('add_webassets_setting', add_setting)
    config.add_directive('add_webassets_path', add_path)

    if assets_env.config['static_view']:
        config.add_static_view(
            settings['webassets.base_url'],
            settings['webassets.base_dir'],
            cache_max_age=assets_env.config['cache_max_age']
        )
        config.add_static_view(
            path.join(assets_env.url, 'webassets-external'),
            path.join(assets_env.directory, 'webassets-external'),
            cache_max_age=assets_env.config['cache_max_age']
        )

    config.add_request_method(get_webassets_env_from_request,
                              'webassets_env', reify=True)
    config.add_request_method(assets, 'webassets', reify=True)



import hashlib
import unittest
import os
import re

from mock import Mock
from pyramid import testing
import pytest
from webassets import __version__ as webassets_version
from webassets.test import TempDirHelper as WebassetsTempDirHelper


def _urls(bundle, env):
    if webassets_version > (0, 9):
        with bundle.bind(env):
            return bundle.urls()
    else:
        return bundle.urls(env=env)


class TempDirHelper(WebassetsTempDirHelper):
    def setup(self):
        super(TempDirHelper, self).setup()
        import sys
        sys.path.append(self.tempdir)

    def teardown(self):
        super(TempDirHelper, self).teardown()

        import sys
        # remove tempdir path elements
        for pth in sys.path:
            if pth.startswith(self.tempdir):
                sys.path.remove(pth)

        sys_modules_items = sys.modules.copy()
        # remove tempdir modules
        for name, module in sys_modules_items.items():
            if module is not None:
                if getattr(module, '__file__', '').startswith(self.tempdir):
                    del sys.modules[name]


class TestWebAssets(unittest.TestCase):
    def test_asset_interface(self):
        from pyramid_webassets import IWebAssetsEnvironment

        def make_env():
            IWebAssetsEnvironment('1')

        self.assertRaises(TypeError, make_env)

    def test_add_web_asset(self):
        from pyramid_webassets import add_webasset

        config = Mock()
        config.registry = Mock()
        queryUtility = Mock()
        env = Mock()
        register = Mock()
        env.register = register
        queryUtility.return_value = env

        config.registry.queryUtility = queryUtility

        add_webasset(config, 'foo', 'bar')
        register.assert_called_with('foo', 'bar')

    def test_add_setting(self):
        from pyramid_webassets import add_setting

        config = Mock()
        config.registry = Mock()
        queryUtility = Mock()
        env = Mock()
        env.config = {}
        register = Mock()
        env.register = register
        queryUtility.return_value = env

        config.registry.queryUtility = queryUtility

        add_setting(config, 'foo', 'bar')
        assert env.config['foo'] == 'bar'

    def test_add_path(self):
        from pyramid_webassets import add_path

        config = Mock()
        config.registry = Mock()
        queryUtility = Mock()
        env = Mock()
        env.config = {}
        register = Mock()
        env.register = register
        queryUtility.return_value = env

        config.registry.queryUtility = queryUtility

        add_path(config, 'foo', '/bar')
        env.append_path.assert_called_with('foo', '/bar')

    def test_get_webassets_env(self):
        from pyramid_webassets import get_webassets_env
        from pyramid_webassets import IWebAssetsEnvironment

        config = Mock()
        config.registry = Mock()
        queryUtility = Mock()

        config.registry.queryUtility = queryUtility

        env = get_webassets_env(config)
        queryUtility.assert_called_with(IWebAssetsEnvironment)

        assert env is not None

    def test_get_webassets_env_from_settings_no_config(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {}

        with self.assertRaises(Exception) as cm:
            get_webassets_env_from_settings(settings)

        assert str(cm.exception) == "You need to provide webassets.base_dir in your configuration"

    def test_get_webassets_env_from_settings_no_base_dir(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {'webassets.base_url': 'static'}

        with self.assertRaises(Exception) as cm:
            get_webassets_env_from_settings(settings)

        assert str(cm.exception) == "You need to provide webassets.base_dir in your configuration"

    def test_get_webassets_env_from_settings_no_base_url(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {'webassets.base_dir': '/home'}

        with self.assertRaises(Exception) as cm:
            get_webassets_env_from_settings(settings)

        assert str(cm.exception) == "You need to provide webassets.base_url in your configuration"

    def test_get_webassets_env_from_settings_minimal(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
        }

        env = get_webassets_env_from_settings(settings)

        assert env.directory == settings['webassets.base_dir']
        assert env.url == '/' + settings['webassets.base_url']
        self.assertEqual(env.url_mapping, {})
        self.assertEqual(env.load_path, [])

    def test_get_webassets_env_from_settings_asset_spec_dir(self):
        import pyramid_webassets
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': 'pyramid_webassets:static',
        }

        env = get_webassets_env_from_settings(settings)
        expected = os.path.join(pyramid_webassets.__path__[0], 'static')

        assert env.directory == expected

    def test_get_webassets_env_from_settings_dir_with_colon(self):
        import pyramid_webassets
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': 'here:static',
        }

        env = get_webassets_env_from_settings(settings)
        expected = os.path.abspath('here:static')

        assert env.directory == expected

    def test_get_webassets_env_from_settings_paths(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.paths': '{"/path/to/scripts": "/js", "/path/to/styles": "/css"}',
        }

        env = get_webassets_env_from_settings(settings)

        assert env.directory == settings['webassets.base_dir']
        assert env.url == '/' + settings['webassets.base_url']
        self.assertEqual(env.url_mapping, {
            '/path/to/scripts': '/js',
            '/path/to/styles':  '/css',
            })
        self.assertEqual(sorted(env.load_path), sorted([
            '/path/to/styles',
            '/path/to/scripts',
            ]))

    def test_get_webassets_env_from_settings_paths_null_url(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.paths': '{"/path/to/scripts": null, "/path/to/styles": "/css"}',
        }

        env = get_webassets_env_from_settings(settings)

        assert env.directory == settings['webassets.base_dir']
        assert env.url == '/' + settings['webassets.base_url']
        self.assertEqual(env.url_mapping, {
            '/path/to/styles':  '/css',
            })
        self.assertEqual(sorted(env.load_path), sorted([
            '/path/to/styles',
            '/path/to/scripts',
            ]))

    def test_get_webassets_env_from_settings_autobuild_disabled(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.auto_build': 'false'
        }

        env = get_webassets_env_from_settings(settings)

        assert env.directory == settings['webassets.base_dir']
        assert env.url == '/' + settings['webassets.base_url']
        assert env.auto_build is False

    def test_get_webassets_env_from_settings_complete(self):
        from pyramid_webassets import get_webassets_env_from_settings
        import webassets

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.debug': 'true',
            'webassets.cache': 'false',
            'webassets.updater': 'always',
            'webassets.jst_compiler': 'Handlebars.compile',
            'webassets.jst_namespace': 'window.Ember.TEMPLATES'
        }

        env = get_webassets_env_from_settings(settings)

        assert env.directory == settings['webassets.base_dir']
        assert env.url == '/' + settings['webassets.base_url']
        assert env.debug is True
        assert isinstance(env.updater, webassets.updater.AlwaysUpdater)
        assert env.config['JST_COMPILER'] == settings['webassets.jst_compiler']
        assert env.config['JST_NAMESPACE'] == settings['webassets.jst_namespace']
        assert env.cache is None
        assert env.auto_build is True

    def test_get_webassets_env_from_settings_with_cache(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.cache': 'true',
        }

        env = get_webassets_env_from_settings(settings)

        assert env.cache is not None

    def test_get_webassets_env_from_settings_with_cache_directory(self):
        from pyramid_webassets import get_webassets_env_from_settings
        from uuid import uuid4
        import shutil

        tmpdir = os.path.join(os.getcwd(), 'test-cache-dir-' + str(uuid4()))

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.cache': tmpdir,
        }

        assert not os.path.isdir(tmpdir)

        env = get_webassets_env_from_settings(settings)

        assert env.cache is not None
        assert env.cache == tmpdir
        assert os.path.isdir(tmpdir)

        shutil.rmtree(tmpdir)

    def test_get_webassets_env_from_settings_prefix_change(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'foo.base_url': 'static',
            'foo.base_dir': os.getcwd(),
        }

        env = get_webassets_env_from_settings(settings, prefix='foo')

        assert env is not None
        assert env.directory == settings['foo.base_dir']
        assert env.url == '/' + settings['foo.base_url']

    def test_get_webassets_env_from_settings_prefix_bad_change(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'foo.base_url': 'static',
            'foo.base_dir': os.getcwd(),
        }

        with self.assertRaises(Exception) as cm:
            get_webassets_env_from_settings(settings, prefix='webassets')

        assert str(cm.exception) == "You need to provide webassets.base_dir in your configuration"

    def test_includeme(self):
        from pyramid_webassets import includeme
        from pyramid_webassets import add_webasset
        from pyramid_webassets import get_webassets_env
        from pyramid_webassets import add_setting
        from pyramid_webassets import get_webassets_env_from_request

        config = Mock()
        add_directive = Mock()
        registerUtility = Mock()
        add_request_method = Mock()

        config.registry = Mock()
        config.registry.registerUtility = registerUtility
        config.add_directive = add_directive
        config.add_request_method = add_request_method

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
        }

        config.registry.settings = settings

        includeme(config)

        expected1 = ('add_webasset', add_webasset)
        expected2 = ('get_webassets_env', get_webassets_env)
        expected3 = ('add_webassets_setting', add_setting)

        assert add_directive.call_args_list[0][0] == expected1
        assert add_directive.call_args_list[1][0] == expected2
        assert add_directive.call_args_list[2][0] == expected3

        assert add_request_method.call_args_list[0][0] == \
            (get_webassets_env_from_request, 'webassets_env')

    def test_get_webassets_env_from_request(self):
        from pyramid_webassets import get_webassets_env_from_request
        from pyramid_webassets import IWebAssetsEnvironment

        request = Mock()
        request.registry = Mock()
        queryUtility = Mock()
        request.registry.queryUtility = queryUtility

        env = get_webassets_env_from_request(request)
        queryUtility.assert_called_with(IWebAssetsEnvironment)

        assert env is not None

    def test_webassets_static_view_setting(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.static_view': True,
        }

        env = get_webassets_env_from_settings(settings)

        assert env is not None
        assert env.config['static_view'] == settings['webassets.static_view']

    def test_webassets_static_view_cache_control_setting(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.static_view': True,
            'webassets.cache_max_age': 3600,
        }

        env = get_webassets_env_from_settings(settings)

        assert env is not None
        assert env.config['static_view'] == settings['webassets.static_view']
        assert env.config['cache_max_age'] == settings['webassets.cache_max_age']

    def test_get_webassets_env_from_settings_load_path(self):
        from pyramid_webassets import get_webassets_env_from_settings

        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.load_path': '/foo bar/\nbaz'
        }

        env = get_webassets_env_from_settings(settings)

        assert env.load_path == ['/foo', 'bar/', 'baz']

    def test_auto_bool(self):
        from pyramid_webassets import get_webassets_env_from_settings
        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.less_run_in_debug': 'true',
        }
        env = get_webassets_env_from_settings(settings)
        assert env.config['less_run_in_debug'] is True

    def test_auto_json(self):
        from pyramid_webassets import get_webassets_env_from_settings
        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.less_extra_args': 'json:["--foo", "--bar"]',
        }
        env = get_webassets_env_from_settings(settings)
        assert env.config['less_extra_args'] == ['--foo', '--bar']


class TestAssetSpecs(TempDirHelper, unittest.TestCase):
    # Mask the methods from TempDirHelper, pytest will try to call them without
    # instantiating the class...
    setup = None
    teardown = None

    def setUp(self):
        from pyramid_webassets import get_webassets_env

        TempDirHelper.setup(self)

        self.create_files({
            'static/__init__.py': '',
            'static/assets/zing.css': '* { text-decoration: underline }'})

        self.request = testing.DummyRequest()
        self.config = testing.setUp(request=self.request, settings={
            'webassets.base_url': 'static',
            'webassets.base_dir': 'static:assets',
            'webassets.static_view': True,
        })
        self.config.include('pyramid_webassets')

        self.env = get_webassets_env(self.config)
        # Disable cache busting
        self.env.url_expire = False

        import sys
        sys.path.append(self.tempdir)

    def tearDown(self):
        TempDirHelper.teardown(self)
        testing.tearDown()

    def test_asset_spec_passthru_uses_static_url(self):
        from webassets import Bundle
        from pyramid.path import AssetResolver

        asset_spec = 'static:assets/zing.css'
        bundle = Bundle(asset_spec)
        self.request.static_url = Mock(return_value='http://example.com/foo/')

        urls = _urls(bundle, self.env)
        path = AssetResolver(None).resolve(asset_spec).abspath()
        self.request.static_url.assert_called_with(path)
        assert urls == ['http://example.com/foo/']

    def test_asset_spec_source_is_resolved(self):
        from webassets import Bundle

        asset_spec = 'static:assets/zing.css'
        bundle = Bundle(asset_spec, output='zung.css')

        urls = _urls(bundle, self.env)
        assert urls == ['http://example.com/static/zung.css']

        fname = os.path.join(self.tempdir, 'static', 'assets', 'zung.css')
        with open(fname) as f:
            assert f.read() == '* { text-decoration: underline }'

    def test_asset_spec_output_is_resolved(self):
        from webassets import Bundle

        asset_spec = 'static:assets/zing.css'
        bundle = Bundle(asset_spec, output='static:assets/zung.css')

        urls = _urls(bundle, self.env)
        assert urls == ['http://example.com/static/zung.css']

        fname = os.path.join(self.tempdir, 'static', 'assets', 'zung.css')
        with open(fname) as f:
            assert f.read() == '* { text-decoration: underline }'

    def test_asset_spec_missing_file(self):
        from webassets import Bundle
        from webassets.exceptions import BundleError

        asset_spec = 'static:assets/bogus.css'
        bundle = Bundle(asset_spec)

        with self.assertRaises(BundleError) as cm:
            _urls(bundle, self.env)

        fname = self.tempdir+'/static/assets/bogus.css'
        assert str(cm.exception) == ("'%s' does not exist" % (fname,))

    def test_asset_spec_missing_package(self):
        from webassets import Bundle
        from webassets.exceptions import BundleError

        asset_spec = 'rabbits:static/zing.css'
        bundle = Bundle(asset_spec)

        with self.assertRaises(BundleError) as cm:
            _urls(bundle, self.env)

        assert re.search('rabbits', str(cm.exception)) is not None

    def test_asset_spec_static_view(self):
        from webassets import Bundle

        asset_spec = 'static:assets/zing.css'
        bundle = Bundle(asset_spec)

        # webassets will copy the file into a place that it can generate
        # a url for
        urls = _urls(bundle, self.env)
        assert self.request.static_url(asset_spec) in urls

    def test_asset_spec_no_static_view(self):
        from webassets import Bundle

        self.create_files({
            'dotted/__init__.py': '',
            'dotted/package/__init__.py': '',
            'dotted/package/name/__init__.py': '',
            'dotted/package/name/static/zing.css':
            '* { text-decoration: underline }'})

        asset_spec = 'dotted.package.name:static/zing.css'
        bundle = Bundle(asset_spec)

        # webassets will copy the file into a place that it can generate
        # a url for
        urls = _urls(bundle, self.env)
        domain = 'http://example.com/static/webassets-external/'

        assert domain in urls[0]
        assert len(urls) == 1

    def test_asset_spec_globbing(self):
        from webassets import Bundle

        self.create_files({
            'static/assets/zang.css':
            '* { text-decoration: underline }'})
        asset_spec = 'static:assets/z*ng.css'
        bundle = Bundle(asset_spec)

        urls = _urls(bundle, self.env)
        assert len(urls) == 2
        assert 'http://example.com/static/zing.css' in urls
        assert 'http://example.com/static/zang.css' in urls

    def test_asset_spec_load_path_and_mapping(self):
        from webassets import Bundle

        asset_path = self.tempdir + '/dotted/package/name/static/'
        self.env.append_path(asset_path, 'http://static.example.com')

        self.create_files({
            'dotted/__init__.py': '',
            'dotted/package/__init__.py': '',
            'dotted/package/name/__init__.py': '',
            'dotted/package/name/static/zing.css':
            '* { text-decoration: underline }'})
        asset_spec = 'dotted.package.name:static/zing.css'
        bundle = Bundle(asset_spec, output=asset_spec.replace('zing', 'zung'))

        urls = _urls(bundle, self.env)
        assert urls == ['http://static.example.com/zung.css']

    def test_bundles_yamlloader_file(self):
        try:
            import yaml
        except ImportError:
            raise unittest.SkipTest('PyYAML not installed')
        from pyramid_webassets import get_webassets_env_from_settings
        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.bundles': self.tempdir + '/foo/bar.yaml',
        }
        self.create_files({'foo/bar.yaml': 'mycss: {contents: style/mycss.css}'})
        env = get_webassets_env_from_settings(settings)
        self.assertEqual(env.config.get('bundles'), None)
        self.assertIsNotNone(env['mycss'])
        self.assertEqual(list(env._named_bundles.keys()), ['mycss'])

    def test_bundles_yamlloader_list(self):
        try:
            import yaml
        except ImportError:
            raise unittest.SkipTest('PyYAML not installed')
        from pyramid_webassets import get_webassets_env_from_settings
        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.bundles': 'dotted.package.name:foo/bar.yaml '
                                 'dotted.package.name:foo/baz.yaml',
        }
        self.create_files({
            'dotted/__init__.py': '',
            'dotted/package/__init__.py': '',
            'dotted/package/name/__init__.py': '',
            'dotted/package/name/foo/bar.yaml': 'mycss: {contents: style/mycss.css}',
            'dotted/package/name/foo/baz.yaml': 'myjs: {contents js/myjs.js}',
        })
        env = get_webassets_env_from_settings(settings)
        self.assertEqual(env.config.get('bundles'), None)
        self.assertIsNotNone(env['mycss'])
        self.assertEqual(sorted(env._named_bundles.keys()), ['mycss', 'myjs'])

    def test_bundles_yamlloader_asset(self):
        try:
            import yaml
        except ImportError:
            raise unittest.SkipTest('PyYAML not installed')
        from pyramid_webassets import get_webassets_env_from_settings
        settings = {
            'webassets.base_url': 'static',
            'webassets.base_dir': os.getcwd(),
            'webassets.bundles': (
                'dotted.package.name:foo/bar.yaml\n'
                'dotted.package.name:foo/baz.yaml'
            ),
        }
        self.create_files({
            'dotted/__init__.py': '',
            'dotted/package/__init__.py': '',
            'dotted/package/name/__init__.py': '',
            'dotted/package/name/foo/bar.yaml': (
                'mycss: {contents: style/mycssoverride.css}'
            ),
            'dotted/package/name/foo/baz.yaml': (
                'mycss: {contents: style/mycss.css}\n'
                'myjs: {contents: js/myjs.js}'
            ),
        })
        env = get_webassets_env_from_settings(settings)
        self.assertEqual(env.config.get('bundles'), None)
        self.assertIsNotNone(env['mycss'])
        self.assertEqual(sorted(env._named_bundles.keys()), ['mycss', 'myjs'])
        self.assertIn('style/mycssoverride.css', env['mycss'].contents)


class TestBaseUrlBehavior(object):
    """
    Tests related to the base_url with asset specs and static_view
    """
    view_actions = {
        None: (False, False),
        'automatic': (True, False),
        'manual': (False, True)}

    def setup_method(self, method):
        self.temp = TempDirHelper()
        self.temp.setup()
        empty_css = '* {color:black}'
        self.temp.create_files({
            'static/__init__.py': '',
            'static/t.css': empty_css,
            'static/some/__init__.py': '',
            'static/some/t.css': empty_css,
            'mypkg/__init__.py': '',
            'mypkg/static/t.css': empty_css})

    def teardown_method(self, method):
        testing.tearDown()
        self.temp.teardown()

    def build_env(self, base_dir, static_view):
        from pyramid_webassets import get_webassets_env

        automatic_view, manual_view = self.view_actions[static_view]

        if ':' not in base_dir:
            base_dir = os.path.join(self.temp.tempdir, base_dir)

        self.request = testing.DummyRequest()
        self.config = testing.setUp(request=self.request, settings={
            'webassets.base_url': 'static',
            'webassets.base_dir': base_dir,
            'webassets.static_view': automatic_view,
            'webassets.url_expire': False,
        })
        self.config.include('pyramid_webassets')

        if manual_view:
            self.config.add_static_view('static', 'mypkg:static')

        return get_webassets_env(self.config)

    def format_expected(self, expected, webasset):
        """
        Formats the expected url when the webasset is external
        """
        from pyramid.path import AssetResolver

        if '%' not in expected:
            return expected

        if ':' in webasset:
            name = AssetResolver(None).resolve(webasset).abspath()
        else:
            name = self.temp.tempdir + '/static/' + webasset

        if webassets_version > (0, 9):
            hashed_filename = hashlib.md5(name.encode("utf-8")).hexdigest()
        else:
            hashed_filename = hash(name) & ((1 << 64) - 1)
        external = 'webassets-external/%s_' % hashed_filename
        return expected % {'external': external}

    @pytest.mark.parametrize(
        ('base_dir', 'static_view', 'webasset', 'expected'),
        [('static', None, 't.css', '/static/t.css'),
         ('static', 'automatic', 't.css', 'http://example.com/static/t.css'),
         ('static', None, 'mypkg:static/t.css', '/static/%(external)st.css'),
         ('static', None, 'static:t.css', '/static/t.css'),

         # If base_dir is an asset spec, but webassets are files
         ('static:some', 'manual', 't.css', '/static/t.css'),
         ('static:some', None, 't.css', '/static/t.css'),

         # base_dir as asset spec, but webasset is asset spec with route
         ('mypkg:some', 'manual', 'mypkg:static/t.css',
          'http://example.com/static/t.css'),

         # base_dir as asset spec, but webasset is asset spec without route
         ('static:some', None, 'static:some/t.css', '/static/t.css'),
         # base_dir as asset spec, but webasset is asset spec automatic route
         ('static:some', 'automatic', 'static:some/t.css',
          'http://example.com/static/t.css'),

         # Work a bit the resolve output (o.css will activate the output=o.css)
         ('mypkg:static', 'manual', 'mypkg:static/t.css',
          'http://example.com/static/o.css'),

         # Test the string joins when using a package root as the asset spec
         ('mypkg:', 'automatic', 'mypkg:static/t.css',
          'http://example.com/static/o.css')],
    )
    def test_url(self, base_dir, static_view, webasset, expected):
        """
        Test final urls.

        If the expected output ends in o.css, it setups output in the bundle.
        """
        from webassets import Bundle

        expected = self.format_expected(expected, webasset)
        params = {} if 'o.css' not in expected else {'output': 'o.css'}
        bundle = Bundle(webasset, **params)
        res = _urls(bundle, self.build_env(base_dir, static_view))
        assert [expected] == res



#!/usr/bin/env python

import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md')) as fp:
    README = fp.read()

with open(os.path.join(here, 'CHANGES.txt')) as fp:
    CHANGES = fp.read()

requires = ['pyramid>=1.4', 'webassets>=0.8', 'zope.interface', 'six>=1.4.1']

extras_require = {
    'bundles-yaml': 'PyYAML>=3.10',
}


class PyTest(TestCommand):
    user_options = [
        ('cov', None, "measure coverage")
    ]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.cov = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['pyramid_webassets/tests']
        if self.cov:
            self.test_args += ['--cov', 'pyramid_webassets']
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

setup(name='pyramid_webassets',
      version='0.10',
      description='pyramid_webassets',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
          "Programming Language :: Python",
          "Framework :: Pylons",
          "License :: OSI Approved :: MIT License",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
      ],
      author='John Anderson',
      author_email='sontek@gmail.com',
      url='http://github.com/sontek/pyramid_webassets',
      keywords='web wsgi bfg pylons pyramid',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='pyramid_webassets',
      install_requires=requires,
      extras_require=extras_require,
      tests_require=['pytest', 'pytest-cov'],
      cmdclass={'test': PyTest},
      )
